'use strict';

// ** Dependencies
const _ = require('lodash');
const path = require('path');
const fs = require('fs-extra');
const errors = require('./errors');
const expand_home_dir = require('expand-home-dir');
const Promise = require('bluebird');
const logger = require('./logging').createLogger('nodus:files');

/**
 * Expands a directory path reference with support for ~ to the home directory.
 * @param filepath
 */
function resolvePath(filepath) {
    filepath = expand_home_dir(filepath);

    // ** Determine the file path using CWD or Absolute path
    if (!path.isAbsolute(filepath)) {
        const cwd = process.cwd();
        filepath = path.resolve(cwd, filepath);
    }

    return filepath;
}

/**
 * Require a file by relative path.
 * @param filepath
 */
function requireFile(filepath) {
    filepath = resolvePath(filepath);

    // ** Ensure the file exists
    if (!fs.existsSync(filepath))
        throw errors('PATH_NOT_FOUND', 'The path to the file could not be found.', {filepath: filepath});

    // ** Load the JSON file
    return require(filepath);
}

function dirname(filepath) {
    return path.dirname(resolvePath(filepath))
}

function filename(filepath) {
    return path.basename(resolvePath(filepath))
}

function isDirectory(path) {
    return fs.lstatSync(path).isDirectory();
}

function isFile(path) {
    return fs.lstatSync(path).isFile();
}

/**
 * Copy a list of files to a specified directory.
 * @param directory
 * @param files
 * @returns {*}
 */
function copyFiles(directory, files) {

    directory = resolvePath(directory);

    if (!isDirectory(directory))
        throw errors('ARGUMENT_ERROR', 'Target must be a directory.', directory);

    return Promise.all(files.map(file => {

        const cwd = process.cwd();

        // Locate the file to copy
        const from = resolvePath(file);

        // Copy the file to the target directory
        const to = path.join(directory, filename(file));

        logger.debug('COPY', {
            from: from,
            to: to
        });

        // Copy the file (async)
        return fs.copy(from, to);
    }));
}

const StreamWriter = file => {

    logger.debug('Creating stream writer for file...', file.filename);

    // Check if the file has a method to write to a stream
    if (file.options.writeStream) {
        logger.debug('File has custom stream writer...', file.filename);
        return file.options.writeStream;
    }

    logger.debug("Using default stream writer...");
    return stream => stream.write(file.contents);
};

class File {
    constructor(filename, contents, options) {

        logger.debug("Creating File...", {filename, options});

        this._filename = filename;
        this._contents = contents;
        this._options = options || {};
    }

    get ClassName() {
        return 'nodus-framework.files.File';
    }

    get filename() {
        return this._filename;
    }

    get extension() {
        return path.extname(this.filename);
    }

    get contents() {
        return this._contents;
    }

    get options() {
        return this._options;
    }

    /**
     * Write the contents of the file to a stream (using .write on contents if available)
     * @param stream
     */
    writeStream(stream) {
        logger.debug('Writing file to stream...', this.filename);

        const writer = StreamWriter(this);
        return writer(stream);
    }
}

// ** Exports
module.exports.File = File;
module.exports.resolvePath = resolvePath;
module.exports.requireFile = requireFile;
module.exports.dirname = dirname;
module.exports.isDirectory = isDirectory;
module.exports.isFile = isFile;
module.exports.copyFiles = copyFiles;