'use strict';

/**
 * A service takes a function and turns it into an object that can handle requests.
 *
 * The service is assigned a log that can be used by the handler via: this.log
 *
 * Request contains the following properties:
 * - id: The unique identifier of the request
 * - args: Arguments to be passed to the request handler
 *
 * Metrics:
 * - Idle Time
 * - Request Rate
 * - Response Time
 * - Total Requests
 * - Total Errors
 *
 * Produces the following Events:
 * - 'request' - Fired when a new request is received.
 * - 'response' - Fired when a response is generated.
 * - 'idle' - Fired when the service is not currently handling any requests.
 * - 'error' - Fired when a request fails.
 */

// ** Dependencies
const _ = require('lodash');
const util = require('util');
const EventEmitter = require('events').EventEmitter;
const activities = require('./activities');
const program = require('./program');
const metrics = require('./metrics');
const microtime = require('./microtime');
const measured = require('measured');
const uuid = require('uuid');
const Promise = require('bluebird');

class Service extends EventEmitter {
    constructor(name, handler, options) {
        super();

        this.name = name;
        this.handler = activities.create(handler);
        this.options = options;

        this._stats = {
            requests: metrics.meter('requests'),
            responses: metrics.histogram('responses'),
            errors: metrics.meter('errors')
        };
    }

    get stats() {
        return _.mapValues(this._stats, stat => stat.toJSON());
    }

    start() {
        this.emit('started');
    }

    stop() {
        this.emit('stopped');
    }

    handle(request) {

        // Assign request an id, if it doesn't already have one
        if (!request.id)
            request.id = uuid.v4();

        // Emit 'request' event
        this.emit('request', request);

        // Track when the request was started
        const started_on = microtime.now();

        // Increment the request-rate metric
        this._stats.requests.mark();

        // Track response time
        return this.handler
            .run(request.args)
            .then(result => {
                const response = {
                    "success": true,
                    "request": request.id,
                    "response": {
                        "data": result
                    }
                };

                this.emit('response', response);
                return response;
            })
            .catch(err => {
                this._stats.errors.mark();

                const response = {
                    "success": false,
                    "request": request.id,
                    "error": err
                };

                this.emit('error', response);
                throw response;
            })
            .finally(() => {
                const completed_on = microtime.now();
                const elapsed_ms = (completed_on - started_on) * 0.001;

                this._stats.responses.update(elapsed_ms);
            })
    }
}

function isService(value) {
    return value instanceof Service;
}

function createService(name, handler, options) {

    if (arguments.length === 1 && util.isFunction(arguments[0]))
        return createService(arguments[0].name, arguments[0]);

    options = options || {};

    const service = new Service(name, handler, options);
    return service;
}

// ** Exports
module.exports = createService;
module.exports.isService = isService;
module.exports.Service = Service;