'use strict';

// ** Dependencies
const WebSocket = require('ws');
const url = require('url');
const os = require('os');
const logger = require('../lib/logging').createLogger('sockets-plugin');

/**
 * Plugin to broadcast all events over a websocket
 * @param server
 * @param service
 */
function loadPlugin(server, service) {

    // Capture hostname
    const service_name = service.name;
    const hostname = os.hostname();

    logger.debug('Creating websockets server...');
    const wss = new WebSocket.Server({server});

    // Add broadcast event to send data to all clients
    wss.broadcast = data => {
        data = JSON.stringify(data);

        wss.clients.forEach(function each(client) {
            if (client.readyState === WebSocket.OPEN) {
                client.send(data);
            }
        });
    };

    wss.on('connection', function connection(ws) {
        const location = url.parse(ws.upgradeReq.url, true);

        logger.debug('CONNECTION', {
            location: location,
            connection: ws
        });

        // You might use location.query.access_token to authenticate or share sessions
        // or ws.upgradeReq.headers.cookie (see http://stackoverflow.com/a/16395220/151312)

        ws.on('message', message => {
            logger.debug('SOCKET_RECEIVED', 'received: %s', message);
        });
    });

    function broadcast_event(event) {

        // Broadcast server events over WebSocket
        service.on(event, function () {

            const payload = {
                timestamp: new Date().toISOString(),
                service: service_name,
                hostname: hostname,
                event: event,
                data: arguments
            };

            // Broadcast to all.
            logger.debug('SOCKET_BROADCAST', payload);
            wss.broadcast(payload);
        });
    }

    broadcast_event('started');
    broadcast_event('stopped');
    broadcast_event('request');
}

// ** Exports
module.exports = loadPlugin;

